// Copyright (c) Microsoft Corporation
// All rights reserved. 
//
// Licensed under the Apache License, Version 2.0 (the ""License""); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 
//
// THIS CODE IS PROVIDED ON AN  *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT. 
//
// See the Apache Version 2.0 License for specific language governing permissions and limitations under the License.

#define NTDDI_VERSION 0x06010000

#include <v8.h>
#include <string>
#include <node_object_wrap.h>
#include <ppltasks.h>
#include "CollectionsConverter.h"
#include "CollectionsWrap.h"
#include "node-async.h"
#include "NodeRtUtils.h"
#include "OpaqueWrapper.h"
#include "WrapperBase.h"

#using <Windows.WinMD>

// this undefs fixes the issues of compiling Windows.Data.Json, Windows.Storag.FileProperties, and Windows.Stroage.Search
// Some of the node header files brings windows definitions with the same names as some of the WinRT methods
#undef DocumentProperties
#undef GetObject
#undef CreateEvent
#undef FindText
#undef SendMessage

const char* REGISTRATION_TOKEN_MAP_PROPERTY_NAME = "__registrationTokenMap__";

using namespace v8;
using namespace node;
using namespace concurrency;

namespace NodeRT { namespace Windows { namespace Globalization { namespace DateTimeFormatting { 
  v8::Handle<v8::Value> WrapDateTimeFormatter(::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ wintRtInstance);
  ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ UnwrapDateTimeFormatter(Handle<Value> value);
  


  static v8::Handle<v8::Value> InitYearFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("YearFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));
    enumObject->Set(String::NewSymbol("abbreviated"), Integer::New(2));
    enumObject->Set(String::NewSymbol("full"), Integer::New(3));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitMonthFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("MonthFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));
    enumObject->Set(String::NewSymbol("abbreviated"), Integer::New(2));
    enumObject->Set(String::NewSymbol("full"), Integer::New(3));
    enumObject->Set(String::NewSymbol("numeric"), Integer::New(4));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitDayOfWeekFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("DayOfWeekFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));
    enumObject->Set(String::NewSymbol("abbreviated"), Integer::New(2));
    enumObject->Set(String::NewSymbol("full"), Integer::New(3));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitDayFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("DayFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitHourFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("HourFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitMinuteFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("MinuteFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitSecondFormatEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("SecondFormat"), enumObject);

    enumObject->Set(String::NewSymbol("none"), Integer::New(0));
    enumObject->Set(String::NewSymbol("default"), Integer::New(1));

    return scope.Close(Undefined());
  }



  
  class DateTimeFormatter : public WrapperBase
  {
  public:    
    static v8::Handle<v8::Value> Init(const Handle<Object> exports)
    {
      HandleScope scope;
      
      s_constructorTemplate = Persistent<FunctionTemplate>::New(FunctionTemplate::New(New));
      s_constructorTemplate->SetClassName(String::NewSymbol("DateTimeFormatter"));
      s_constructorTemplate->InstanceTemplate()->SetInternalFieldCount(1);
      
            
      s_constructorTemplate->PrototypeTemplate()->Set(String::NewSymbol("format"), FunctionTemplate::New(Format)->GetFunction());
      
                        
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("numeralSystem"), NumeralSystemGetter, NumeralSystemSetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("calendar"), CalendarGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("clock"), ClockGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("geographicRegion"), GeographicRegionGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeDay"), IncludeDayGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeDayOfWeek"), IncludeDayOfWeekGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeHour"), IncludeHourGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeMinute"), IncludeMinuteGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeMonth"), IncludeMonthGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeSecond"), IncludeSecondGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("includeYear"), IncludeYearGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("languages"), LanguagesGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("patterns"), PatternsGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("resolvedGeographicRegion"), ResolvedGeographicRegionGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("resolvedLanguage"), ResolvedLanguageGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("template"), TemplateGetter);
      
      Local<Function> constructor = s_constructorTemplate->GetFunction();

      constructor->SetAccessor(String::NewSymbol("longDate"), LongDateGetter);
      constructor->SetAccessor(String::NewSymbol("longTime"), LongTimeGetter);
      constructor->SetAccessor(String::NewSymbol("shortDate"), ShortDateGetter);
      constructor->SetAccessor(String::NewSymbol("shortTime"), ShortTimeGetter);

      exports->Set(String::NewSymbol("DateTimeFormatter"), constructor);
      return scope.Close(Undefined());
    }


    virtual ::Platform::Object^ GetObjectInstance() const override
    {
      return _instance;
    }

  private:
    
    DateTimeFormatter(::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ instance)
    {
      _instance = instance;
    }
    
    
    static v8::Handle<v8::Value> New(const v8::Arguments& args)
    {
      HandleScope scope;

      // in case the constructor was called without the new operator
      if (!s_constructorTemplate->HasInstance(args.This()))
      {
        if (args.Length() > 0)
        {
          std::unique_ptr<Handle<Value> []> constructorArgs(new Handle<Value>[args.Length()]);

          Handle<Value> *argsPtr = constructorArgs.get();
          for (int i = 0; i < args.Length(); i++)
          {
            argsPtr[i] = args[i];
          }

          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), constructorArgs.get());
        }
        else
        {
          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), nullptr);
        }
      }
      
      ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ winRtInstance;


      if (args.Length() == 1 && OpaqueWrapper::IsOpaqueWrapper(args[0]) &&
        NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(args[0]))
      {
        try 
        {
          winRtInstance = (::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^) NodeRT::Utils::GetObjectInstance(args[0]);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 1
        && args[0]->IsString())
      {
        try
        {
          Platform::String^ arg0 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[0])));
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 2
        && args[0]->IsString()
        && (NodeRT::Utils::IsWinRtWrapperOf<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(args[1]) || args[1]->IsArray()))
      {
        try
        {
          Platform::String^ arg0 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[0])));
          ::Windows::Foundation::Collections::IIterable<::Platform::String^>^ arg1 = 
            [] (v8::Handle<v8::Value> value) -> ::Windows::Foundation::Collections::IIterable<::Platform::String^>^
            {
              if (value->IsArray())
              {
                return NodeRT::Collections::JsArrayToWinrtVector<::Platform::String^>(value.As<Array>(), 
                 [](Handle<Value> value) -> bool {
                   return (!NodeRT::Utils::IsWinRtWrapper(value));
                 },
                 [](Handle<Value> value) -> ::Platform::String^ {
                   return ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));
                 }
                );
              }
              else
              {
                return dynamic_cast<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(NodeRT::Utils::GetObjectInstance(value));
              }
            } (args[1]);
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0,arg1);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 5
        && args[0]->IsString()
        && (NodeRT::Utils::IsWinRtWrapperOf<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(args[1]) || args[1]->IsArray())
        && args[2]->IsString()
        && args[3]->IsString()
        && args[4]->IsString())
      {
        try
        {
          Platform::String^ arg0 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[0])));
          ::Windows::Foundation::Collections::IIterable<::Platform::String^>^ arg1 = 
            [] (v8::Handle<v8::Value> value) -> ::Windows::Foundation::Collections::IIterable<::Platform::String^>^
            {
              if (value->IsArray())
              {
                return NodeRT::Collections::JsArrayToWinrtVector<::Platform::String^>(value.As<Array>(), 
                 [](Handle<Value> value) -> bool {
                   return (!NodeRT::Utils::IsWinRtWrapper(value));
                 },
                 [](Handle<Value> value) -> ::Platform::String^ {
                   return ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));
                 }
                );
              }
              else
              {
                return dynamic_cast<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(NodeRT::Utils::GetObjectInstance(value));
              }
            } (args[1]);
          Platform::String^ arg2 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[2])));
          Platform::String^ arg3 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[3])));
          Platform::String^ arg4 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[4])));
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0,arg1,arg2,arg3,arg4);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 4
        && args[0]->IsInt32()
        && args[1]->IsInt32()
        && args[2]->IsInt32()
        && args[3]->IsInt32())
      {
        try
        {
          ::Windows::Globalization::DateTimeFormatting::YearFormat arg0 = static_cast<::Windows::Globalization::DateTimeFormatting::YearFormat>(args[0]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::MonthFormat arg1 = static_cast<::Windows::Globalization::DateTimeFormatting::MonthFormat>(args[1]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::DayFormat arg2 = static_cast<::Windows::Globalization::DateTimeFormatting::DayFormat>(args[2]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat arg3 = static_cast<::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat>(args[3]->Int32Value());
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0,arg1,arg2,arg3);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 3
        && args[0]->IsInt32()
        && args[1]->IsInt32()
        && args[2]->IsInt32())
      {
        try
        {
          ::Windows::Globalization::DateTimeFormatting::HourFormat arg0 = static_cast<::Windows::Globalization::DateTimeFormatting::HourFormat>(args[0]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::MinuteFormat arg1 = static_cast<::Windows::Globalization::DateTimeFormatting::MinuteFormat>(args[1]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::SecondFormat arg2 = static_cast<::Windows::Globalization::DateTimeFormatting::SecondFormat>(args[2]->Int32Value());
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0,arg1,arg2);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 8
        && args[0]->IsInt32()
        && args[1]->IsInt32()
        && args[2]->IsInt32()
        && args[3]->IsInt32()
        && args[4]->IsInt32()
        && args[5]->IsInt32()
        && args[6]->IsInt32()
        && (NodeRT::Utils::IsWinRtWrapperOf<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(args[7]) || args[7]->IsArray()))
      {
        try
        {
          ::Windows::Globalization::DateTimeFormatting::YearFormat arg0 = static_cast<::Windows::Globalization::DateTimeFormatting::YearFormat>(args[0]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::MonthFormat arg1 = static_cast<::Windows::Globalization::DateTimeFormatting::MonthFormat>(args[1]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::DayFormat arg2 = static_cast<::Windows::Globalization::DateTimeFormatting::DayFormat>(args[2]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat arg3 = static_cast<::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat>(args[3]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::HourFormat arg4 = static_cast<::Windows::Globalization::DateTimeFormatting::HourFormat>(args[4]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::MinuteFormat arg5 = static_cast<::Windows::Globalization::DateTimeFormatting::MinuteFormat>(args[5]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::SecondFormat arg6 = static_cast<::Windows::Globalization::DateTimeFormatting::SecondFormat>(args[6]->Int32Value());
          ::Windows::Foundation::Collections::IIterable<::Platform::String^>^ arg7 = 
            [] (v8::Handle<v8::Value> value) -> ::Windows::Foundation::Collections::IIterable<::Platform::String^>^
            {
              if (value->IsArray())
              {
                return NodeRT::Collections::JsArrayToWinrtVector<::Platform::String^>(value.As<Array>(), 
                 [](Handle<Value> value) -> bool {
                   return (!NodeRT::Utils::IsWinRtWrapper(value));
                 },
                 [](Handle<Value> value) -> ::Platform::String^ {
                   return ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));
                 }
                );
              }
              else
              {
                return dynamic_cast<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(NodeRT::Utils::GetObjectInstance(value));
              }
            } (args[7]);
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0,arg1,arg2,arg3,arg4,arg5,arg6,arg7);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 11
        && args[0]->IsInt32()
        && args[1]->IsInt32()
        && args[2]->IsInt32()
        && args[3]->IsInt32()
        && args[4]->IsInt32()
        && args[5]->IsInt32()
        && args[6]->IsInt32()
        && (NodeRT::Utils::IsWinRtWrapperOf<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(args[7]) || args[7]->IsArray())
        && args[8]->IsString()
        && args[9]->IsString()
        && args[10]->IsString())
      {
        try
        {
          ::Windows::Globalization::DateTimeFormatting::YearFormat arg0 = static_cast<::Windows::Globalization::DateTimeFormatting::YearFormat>(args[0]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::MonthFormat arg1 = static_cast<::Windows::Globalization::DateTimeFormatting::MonthFormat>(args[1]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::DayFormat arg2 = static_cast<::Windows::Globalization::DateTimeFormatting::DayFormat>(args[2]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat arg3 = static_cast<::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat>(args[3]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::HourFormat arg4 = static_cast<::Windows::Globalization::DateTimeFormatting::HourFormat>(args[4]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::MinuteFormat arg5 = static_cast<::Windows::Globalization::DateTimeFormatting::MinuteFormat>(args[5]->Int32Value());
          ::Windows::Globalization::DateTimeFormatting::SecondFormat arg6 = static_cast<::Windows::Globalization::DateTimeFormatting::SecondFormat>(args[6]->Int32Value());
          ::Windows::Foundation::Collections::IIterable<::Platform::String^>^ arg7 = 
            [] (v8::Handle<v8::Value> value) -> ::Windows::Foundation::Collections::IIterable<::Platform::String^>^
            {
              if (value->IsArray())
              {
                return NodeRT::Collections::JsArrayToWinrtVector<::Platform::String^>(value.As<Array>(), 
                 [](Handle<Value> value) -> bool {
                   return (!NodeRT::Utils::IsWinRtWrapper(value));
                 },
                 [](Handle<Value> value) -> ::Platform::String^ {
                   return ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));
                 }
                );
              }
              else
              {
                return dynamic_cast<::Windows::Foundation::Collections::IIterable<::Platform::String^>^>(NodeRT::Utils::GetObjectInstance(value));
              }
            } (args[7]);
          Platform::String^ arg8 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[8])));
          Platform::String^ arg9 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[9])));
          Platform::String^ arg10 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[10])));
          
          winRtInstance = ref new ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter(arg0,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9,arg10);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Invalid arguments, no suitable constructor found")));
        return scope.Close(Undefined());
      }

      args.This()->SetHiddenValue(String::NewSymbol("__winRtInstance__"), True());

      DateTimeFormatter *wrapperInstance = new DateTimeFormatter(winRtInstance);
      wrapperInstance->Wrap(args.This());

      return args.This();
    }


  
    static Handle<Value> Format(const v8::Arguments& args)
    {
      HandleScope scope;

      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(args.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(args.This());

      if (args.Length() == 1
        && args[0]->IsDate())
      {
        try
        {
          ::Windows::Foundation::DateTime arg0 = NodeRT::Utils::DateTimeFromJSDate(args[0]);
          
          Platform::String^ result;
          result = wrapper->_instance->Format(arg0);
          return scope.Close(NodeRT::Utils::NewString(result->Data()));
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 2
        && args[0]->IsDate()
        && args[1]->IsString())
      {
        try
        {
          ::Windows::Foundation::DateTime arg0 = NodeRT::Utils::DateTimeFromJSDate(args[0]);
          Platform::String^ arg1 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[1])));
          
          Platform::String^ result;
          result = wrapper->_instance->Format(arg0, arg1);
          return scope.Close(NodeRT::Utils::NewString(result->Data()));
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else 
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: no suitable overload found")));
        return scope.Close(Undefined());
      }

      return scope.Close(Undefined());
    }



    static Handle<Value> NumeralSystemGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->NumeralSystem;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static void NumeralSystemSetter(Local<String> property, Local<Value> value, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!value->IsString())
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Value to set is of unexpected type")));
        return;
      }

      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return;
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        
        Platform::String^ winRtValue = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));

        wrapper->_instance->NumeralSystem = winRtValue;
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
      }
    }
    
    static Handle<Value> CalendarGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->Calendar;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ClockGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->Clock;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> GeographicRegionGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->GeographicRegion;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeDayGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::DayFormat result = wrapper->_instance->IncludeDay;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeDayOfWeekGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::DayOfWeekFormat result = wrapper->_instance->IncludeDayOfWeek;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeHourGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::HourFormat result = wrapper->_instance->IncludeHour;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeMinuteGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::MinuteFormat result = wrapper->_instance->IncludeMinute;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeMonthGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::MonthFormat result = wrapper->_instance->IncludeMonth;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeSecondGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::SecondFormat result = wrapper->_instance->IncludeSecond;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> IncludeYearGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::YearFormat result = wrapper->_instance->IncludeYear;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> LanguagesGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Foundation::Collections::IVectorView<::Platform::String^>^ result = wrapper->_instance->Languages;
        return scope.Close(NodeRT::Collections::VectorViewWrapper<::Platform::String^>::CreateVectorViewWrapper(result, 
            [](::Platform::String^ val) -> Handle<Value> {
              return NodeRT::Utils::NewString(val->Data());
            },
            [](Handle<Value> value) -> bool {
              return value->IsString();
            },
            [](Handle<Value> value) -> ::Platform::String^ {
              return ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));
            }
          ));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> PatternsGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        ::Windows::Foundation::Collections::IVectorView<::Platform::String^>^ result = wrapper->_instance->Patterns;
        return scope.Close(NodeRT::Collections::VectorViewWrapper<::Platform::String^>::CreateVectorViewWrapper(result, 
            [](::Platform::String^ val) -> Handle<Value> {
              return NodeRT::Utils::NewString(val->Data());
            },
            [](Handle<Value> value) -> bool {
              return value->IsString();
            },
            [](Handle<Value> value) -> ::Platform::String^ {
              return ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(value)));
            }
          ));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ResolvedGeographicRegionGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->ResolvedGeographicRegion;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ResolvedLanguageGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->ResolvedLanguage;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> TemplateGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      DateTimeFormatter *wrapper = DateTimeFormatter::Unwrap<DateTimeFormatter>(info.This());

      try 
      {
        Platform::String^ result = wrapper->_instance->Template;
        return scope.Close(NodeRT::Utils::NewString(result->Data()));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    

    static Handle<Value> LongDateGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ result = ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter::LongDate;
        return scope.Close(WrapDateTimeFormatter(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> LongTimeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ result = ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter::LongTime;
        return scope.Close(WrapDateTimeFormatter(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ShortDateGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ result = ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter::ShortDate;
        return scope.Close(WrapDateTimeFormatter(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ShortTimeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;

      try 
      {
        ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ result = ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter::ShortTime;
        return scope.Close(WrapDateTimeFormatter(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    

  private:
    ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ _instance;
    static Persistent<FunctionTemplate> s_constructorTemplate;

    friend v8::Handle<v8::Value> WrapDateTimeFormatter(::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ wintRtInstance);
    friend ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ UnwrapDateTimeFormatter(Handle<Value> value);
    friend bool IsDateTimeFormatterWrapper(Handle<Value> value);
  };
  Persistent<FunctionTemplate> DateTimeFormatter::s_constructorTemplate;

  v8::Handle<v8::Value> WrapDateTimeFormatter(::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ winRtInstance)
  {
    HandleScope scope;

    if (winRtInstance == nullptr)
    {
      return scope.Close(Undefined());
    }

    Handle<Object> opaqueWrapper = CreateOpaqueWrapper(winRtInstance);
    Handle<Value> args[] = {opaqueWrapper};
    return scope.Close(DateTimeFormatter::s_constructorTemplate->GetFunction()->NewInstance(_countof(args), args));
  }

  ::Windows::Globalization::DateTimeFormatting::DateTimeFormatter^ UnwrapDateTimeFormatter(Handle<Value> value)
  {
     return DateTimeFormatter::Unwrap<DateTimeFormatter>(value.As<Object>())->_instance;
  }

  void InitDateTimeFormatter(Handle<Object> exports)
  {
    DateTimeFormatter::Init(exports);
  }

} } } } 

void init(Handle<Object> exports)
{
  if (FAILED(CoInitializeEx(nullptr, COINIT_MULTITHREADED)))
  {
    ThrowException(v8::Exception::Error(NodeRT::Utils::NewString(L"error in CoInitializeEx()")));
    return;
  }
  
  NodeRT::Windows::Globalization::DateTimeFormatting::InitYearFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitMonthFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitDayOfWeekFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitDayFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitHourFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitMinuteFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitSecondFormatEnum(exports);
  NodeRT::Windows::Globalization::DateTimeFormatting::InitDateTimeFormatter(exports);

  NodeRT::Utils::RegisterNameSpace("Windows.Globalization.DateTimeFormatting", exports);
}


NODE_MODULE(binding, init)